*! LinReg version 1.0.0

version 15
set matastrict on

mata:

class LinReg
{
        // --------------------------------------- public functions 

    public:
        void                   setup()  // sets model to be fitted
        real scalar            N()      // # obs 
        real scalar            k()      // # X vars excluding cons
        real scalar            K()      // # X vars including cons

        real colvector         b()      // coefs; K x 1

        real matrix            V()      // variance matrix
        real scalar            s2()     // variance of e
        real scalar            ee()     // e'e, error sum of squares

        real colvector         yhat()   // predicted values

        real scalar            k_adj()  // k - (# dropped due to collinearity)
        real scalar            K_adj () // K - (# dropped due to collinearity)

        real rowvector         mean_X() // mean(X), 1 x k
        real scalar            mean_y() // mean(y), 1 x 1

        // --------------------------------------- private functions 

    private:
        void clear()                    // clear contents of r

        real colvector b_X()            // coefs excluding intercept; k x 1
        real scalar    b_c()            // intercept, 1 x 1 (0 if nocons)

        real matrix    XX()             // X'X, K x K
        real colvector Xy()             // X'y  K x 1
        real matrix    XXinv()          // (X'X)^(-1), K x K
        real scalar    yy()             // y'y

        /* Deviation-from-means notation: 

                S = X :- mean(X)    S: N x k
                t = y := mean(y)    t: N x 1

           This is notation; S and T are not physically built. 
        */
        
        real matrix    SS()             // S'S, k x k 
        real colvector St()             // S't, K x 1 (sic)
        real matrix    SSinv()          // (S'S)^(-1), K x K (sic)
        void calc_SSinv_cons()          // subroutine of SSinv()


        // --------------------- self-threading code input variables
    private:
        pointer(real colvector) scalar  y     
        pointer(real matrix) scalar     X     
        real scalar                     cons  

        // ------------------- self-threading code derived variables

    private:
        real colvector  b
        real matrix     V
        real scalar     s2
        real scalar     yy
        real scalar     ee
        real rowvector  mean_X
        real scalar     mean_y
        real scalar     K_adj
        real scalar     k_adj

        // non deviated matrices, used if no intercept:
        real matrix    XX
        real colvector Xy
        real matrix    XXinv

        // deviated matrices, used if intercept:
        real matrix    SS
        real matrix    St
        real matrix    SSinv
}
                                

void LinReg::clear()
{
        y       = NULL
        X       = NULL

        XX      = .z
        Xy      = .z
        XXinv   = .z 
        b       = .z
        V       = .z
        s2      = .z
        yy      = .z
        ee      = .z

        SS      = .z
        St      = .z
        mean_X  = .z
        mean_y  = .z
        SSinv   = .z
        K_adj   = .z 
        k_adj   = .z 
}


/* -----------------------------------------------public functions --- */

void LinReg::setup(real colvector user_y, 
                   real matrix    user_X,
                   real scalar    user_cons )
{
        assert(rows(user_y)==rows(user_X))

        clear()
        y    = &user_y
        X    = &user_X
        cons = (user_cons!=0)
}

real scalar LinReg::N() return(rows(*X))
real scalar LinReg::k() return(cols(*X))
real scalar LinReg::K() return(cols(*X) + cons)

/*
        Throughout this code, distinguish carefully between k and K:

                k = # of independent variables EXCLUDING intercept
                K = # of independent variables INCLUDING intercept

        In models without an intercept, K = k
        In models with    an intercept, K = k + 1

        Note: *r.X is k x k, X does not include vector of 1s, and 
                 b is K x 1, coefficient include intercept if there is one.

        To code the equivalent of X*b, you extract the first k elements
        of b and add the intercept separately.  To make this easier
        think of

                 b = (b_X, b_c)   

        b_X() returns b_X:  1 x k
        b_c() returns b_c:  1 x 1  containing b[K] or 0. 

        Thus, X*b can be calculated as (*X)*b_X() :+ b_c() 
*/


real colvector LinReg::b_X()
{
        return(cons ? ( k() ? b()[|1\k()|] : J(0, 1, .) ) : b())
}


real scalar LinReg::b_c() return(cons ? b()[K()] : 0) 


real colvector LinReg::b()
{
        if (b == .z) { 
                if (cons) {
                        b      = SSinv() * St()
                        b[K()] = mean_y() - mean_X()*b_X()

                }
                else    b      = XXinv() * Xy()
        }
        return(b)
}


real matrix LinReg::V()
{
        if (V == .z) V = s2() * (cons ? SSinv() : XXinv())
        return(V)
}


real scalar LinReg::s2()
{
        if (s2 == .z) s2 = ee() / ( N() - K_adj() )
        return(s2)
}


real scalar LinReg::ee()
{
        real colvector   e

        if (ee == .z) {
                e  = *y - yhat()
                ee = quadcross(e, e)
        }
        return(ee)
}


real colvector LinReg::yhat()
        return(cons ? (*X)*b_X() :+ b_c() : (*X)*b() )


/* ------------------------------------------internal subroutines --- */

real matrix LinReg::XX()
{
        if (XX == .z) XX = cross(*X, cons,  *X, cons)
        return(XX)
}


real colvector LinReg::Xy()
{
        if (Xy == .z) Xy = cross(*X, cons,  *y, 0)
        return(Xy)
}


real scalar LinReg::yy()
{
        if (yy == .z) yy = cross(*y, *y)
        return(yy)
}


real matrix LinReg::XXinv()
{
        if (XXinv == .z) XXinv = invsym(XX())
        return(XXinv)
}


real matrix LinReg::SSinv()
{
        if (SSinv == .z) {
                if (cons)  calc_SSinv_cons()
                else      _error("SSinv() invalid if cons==0")
        }
        return(SSinv)
}


void LinReg::calc_SSinv_cons()
{
        real scalar     K, k, i
        real rowvector  v
        real matrix     kxk, rowK, colK

        K = K()              // # X vars including constant
        k = k()              // # x vars excluding constant

        /*
            Calculate inverse of S'S = SS().
            This routine called only if cons, meaning K = k + 1.
            SS() is k x k -- it excludes constant.
            We want the full K x K inverse.
        */

        // ---------------------------------------- SSinv 1 x 1 case ---
        if (K==1) {
                SSinv = 1/N()
                return
        }

        // ----------------------------------- SSinv 2 x 2 or larger ---

        SSinv = J(K, K, .)  // matrix now K x K

                             // Range subscripts:
        kxk  = (1,1 \ k,k)   //   k x k submatrix  rows & cols 1..k
        rowK = (K,1 \ K,k)   //   1 x k submatrix; row K
        colK = (1,K \ k,K)   //   k x 1 submatrix; col K

        SSinv[|kxk |]     = invsym(SS())

        SSinv[|rowK|] = v = -(mean_X() * SSinv[|kxk|])
        SSinv[|colK|] = v'

        SSinv[ K,K  ]     = 1/N() - mean_X()*SSinv[|colK|]
}




real matrix LinReg::SS()
{
        if (SS == .z) { 
                if (cons) {
                        SS = quadcrossdev(*X, 0, mean_X(), 
                                          *X, 0, mean_X() )
                }
                else         _error("SS() invalid if cons==0")
        }
        return(SS)
}


real colvector LinReg::St()
{
        if (St == .z) {
                if (!cons) _error("St() invalid if cons==0")

                St = quadcrossdev(*X,      0, mean_X(), 
                                  *y,      0, mean_y()) \ 0
        }
        return(St)
}


real rowvector LinReg::mean_X()
{
        if (mean_X == .z) mean_X = mean(*X)
        return(mean_X)
}


real scalar LinReg::mean_y()
{
        if (mean_y == .z) mean_y = mean(*y)
        return(mean_y)
}


real scalar LinReg::K_adj()
{
        if (K_adj == .z) {
                if (cons) K_adj = colsum(diagonal(SSinv()) :!= 0)
                else      K_adj = colsum(diagonal(XXinv()) :!= 0)
        }
        return(K_adj)
}


real scalar LinReg::k_adj()
{
        if (k_adj == .z) k_adj = k() - ( K()-K_adj() )
        return(k_adj) 
}

end
