*! lr*() version 1.0.0

version 15
set matastrict on

mata:

struct lrinfo 
{
        // --------------------------------------- inputs
        pointer(real colvector) scalar  y     
        pointer(real matrix) scalar     X     
        real scalar                     cons  

        // --------------------------------------- derived
        /*
            Notation:
                 y = Xb + e
              yhat = Xb

                 N = # of obs
                 K = # of indep vars including intercept if present
                 k = # of indep vars excluding intercept

             K_adj = K adjusted for collinearity
        */
            
        real matrix     XX          // X'X; K x K 
        real colvector  Xy          // X'y; K x 1
        real matrix     XXinv       // (X'X)^(-1);   K x K 
        real colvector  b           // coefficients; K x 1
        real matrix     V           // variance matrix (VCE); K x K
        real scalar     s2          // variance of e
        real scalar     yy          // y'y
        real scalar     ee          // e'e, error sum of squares
        real scalar     K_adj

        // ------------------------ derived but not stored
}

void lrinit(struct lrinfo scalar r)
{
        r.XX      = .z
        r.Xy      = .z
        r.XXinv   = .z 
        r.b       = .z
        r.V       = .z
        r.s2      = .z
        r.yy      = .z
        r.ee      = .z
        r.K_adj   = .z
}


/* -------------------------------------------- entry points --- */

struct lrinfo scalar lrset( real colvector y, 
                            real matrix    X,  
                            real scalar    cons )
{
        struct lrinfo scalar   r

        assert(rows(y)==rows(X))

        r.y    = &y
        r.X    = &X
        r.cons = (cons!=0)

        lrinit(r)              // initialize derived vars

        return(r) 
}

real scalar lr_N(struct lrinfo scalar r) return(rows(*r.X))
real scalar lr_k(struct lrinfo scalar r) return(cols(*r.X))
real scalar lr_K(struct lrinfo scalar r) return(cols(*r.X) + r.cons)


real scalar lr_K_adj(struct lrinfo scalar r)
{
        if (r.K_adj == .z) {
                r.K_adj = colsum(diagonal(lr_XXinv(r)) :!= 0)
        }
        return(r.K_adj)
}


real colvector lr_b(struct lrinfo scalar r)
{
        if (r.b == .z) {
                r.b = lr_XXinv(r) * lr_Xy(r)
        }
        return(r.b)
}

real matrix lr_V(struct lrinfo scalar r)
{
        if (r.V == .z) {
                r.V = lr_s2(r) * lr_XXinv(r)
        }
        return(r.V)
}

real scalar lr_s2(struct lrinfo scalar r)
{
        if (r.s2 == .z) {
                r.s2 = lr_ee(r) / ( lr_N(r) - lr_K_adj(r) )
        }
        return(r.s2)
}

real scalar lr_ee(struct lrinfo scalar r)
{
        if (r.ee == .z) {
                r.ee = lr_yy(r) - lr_b(r)'lr_XX(r)*lr_b(r)
        }
        return(r.ee)
}


real colvector lr_yhat(struct lrinfo scalar r)
{
        return( (*r.X)*lr_b_X(r) :+ lr_b_c(r) )
}

real colvector lr_b_X(struct lrinfo scalar r)
{
        if (r.cons) {         
                if (lr_k(r)) return( lr_b(r)[| 1 \ lr_k(r) |])
                else         return( J(0, 1, .))
        }
        else                 return(lr_b(r))
}

real scalar lr_b_c(struct lrinfo scalar r)
{
        return( r.cons ? lr_b(r)[lr_K(r)] : 0 )
}


/* -------------------------------------internal subroutines --- */

real matrix lr_XX(struct lrinfo scalar r)
{
        if (r.XX == .z) {
                r.XX = cross(*r.X, r.cons,  *r.X, r.cons)
        }
        return(r.XX)
}

real colvector lr_Xy(struct lrinfo scalar r)
{
        if (r.Xy == .z) {
                r.Xy = cross(*r.X, r.cons,  *r.y, 0)
        }
        return(r.Xy)
}

real scalar lr_yy(struct lrinfo scalar r)
{
        if (r.yy == .z) {
                r.yy = cross(*r.y, *r.y)
        }
        return(r.yy)
}


real matrix lr_XXinv(struct lrinfo scalar r)
{
        if (r.XXinv == .z) {
                r.XXinv = invsym(lr_XX(r))
        }
        return(r.XXinv)
}

end
